/* dazukofs: access control stackable filesystem

   Copyright (C) 2008 John Ogness
     Author: John Ogness <dazukocode@ogness.net>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <linux/device.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>
#include <asm/uaccess.h>

#include "dazukofs_fs.h"
#include "dev.h"

struct dazukofs_proc {
	struct list_head list;
	struct task_struct *curr;
};

static struct dazukofs_proc ign_list;
static struct mutex ign_list_mutex;
static struct kmem_cache *dazukofs_ign_cachep;

int dazukofs_check_ignore_process(void)
{
	struct list_head *pos;
	struct dazukofs_proc *proc;
	int found = 0;

	mutex_lock(&ign_list_mutex);
	list_for_each(pos, &ign_list.list) {
		proc = list_entry(pos, struct dazukofs_proc, list);
		if (proc->curr == current) {
			found = 1;
			break;
		}
	}
	mutex_unlock(&ign_list_mutex);

	return !found;
}

static int dazukofs_add_ign(struct file *file)
{
	struct dazukofs_proc *proc;

	proc = kmem_cache_alloc(dazukofs_ign_cachep, GFP_KERNEL);
	if (!proc) {
		file->private_data = NULL;
		return -ENOMEM;
	}
	memset(proc, 0, sizeof(*proc));

	file->private_data = proc;
	proc->curr = current;

	mutex_lock(&ign_list_mutex);
	list_add(&proc->list, &ign_list.list);
	mutex_unlock(&ign_list_mutex);

	return 0;
}

static void dazukofs_remove_ign(struct file *file)
{
	struct list_head *pos;
	struct dazukofs_proc *proc = NULL;
	struct dazukofs_proc *check_proc = file->private_data;
	int found = 0;

	if (!check_proc)
		return;

	mutex_lock(&ign_list_mutex);
	list_for_each(pos, &ign_list.list) {
		proc = list_entry(pos, struct dazukofs_proc, list);
		if (proc->curr == check_proc->curr) {
			found = 1;
			list_del(pos);
			break;
		}
	}
	mutex_unlock(&ign_list_mutex);

	if (found) {
		file->private_data = NULL;
		kmem_cache_free(dazukofs_ign_cachep, proc);
	}
}

static int dazukofs_ign_open(struct inode *inode, struct file *file)
{
	return dazukofs_add_ign(file);
}

static int dazukofs_ign_release(struct inode *inode, struct file *file)
{
	dazukofs_remove_ign(file);
	return 0;
}

static void dazukofs_destroy_ignlist(void)
{
	struct list_head *pos;
	struct list_head *q;
	struct dazukofs_proc *proc;

	list_for_each_safe(pos, q, &ign_list.list) {
		proc = list_entry(pos, struct dazukofs_proc, list);
		list_del(pos);
		kmem_cache_free(dazukofs_ign_cachep, proc);
	}
}

static struct {
	struct cdev cdev;
	struct device dev;
	struct class_device cls_dev;
} ign_dev;

static struct file_operations ign_fops = {
	.owner		= THIS_MODULE,
	.open		= dazukofs_ign_open,
	.release	= dazukofs_ign_release,
};


static void dazukofs_ign_class_dev_release(struct class_device *dev)
{
	// nothing
}

static void dazukofs_ign_dev_release(struct device *dev)
{
	// nothing
}


int dazukofs_ign_dev_init(int dev_major, int dev_minor,
			  struct class *dazukofs_class)
{
	dev_t devt;
	int err;

	INIT_LIST_HEAD(&ign_list.list);
	mutex_init(&ign_list_mutex);

	dazukofs_ign_cachep =
		kmem_cache_create("dazukofs_ign_cache",
				  sizeof(struct dazukofs_proc), 0,
				  SLAB_HWCACHE_ALIGN, NULL, NULL);
	if (!dazukofs_ign_cachep) 
		return -ENOMEM;

	devt = MKDEV(dev_major, dev_minor);

	/* setup cdev for ignore */
	cdev_init(&ign_dev.cdev, &ign_fops);
	ign_dev.cdev.owner = THIS_MODULE;
	err = cdev_add(&ign_dev.cdev, devt, 1);
	if (err)
		goto error_out;		

	ign_dev.dev.parent = &platform_bus;
	ign_dev.dev.release = dazukofs_ign_dev_release;
	snprintf(ign_dev.dev.bus_id, BUS_ID_SIZE, "%s.ign", DEVICE_NAME);

	err = device_register(&ign_dev.dev);

	if (err) 
		goto error_out2;

	ign_dev.cls_dev.release = dazukofs_ign_class_dev_release;
	ign_dev.cls_dev.class = dazukofs_class;
	ign_dev.cls_dev.dev = &ign_dev.dev;
	snprintf(ign_dev.cls_dev.class_id, BUS_ID_SIZE, "%s.ign", DEVICE_NAME);
	ign_dev.cls_dev.devt = devt;

	err = class_device_register(&ign_dev.cls_dev);

	if (err)
		goto error_out3;


	return 0;
error_out3:
	device_unregister(&ign_dev.dev);
error_out2:
	cdev_del(&ign_dev.cdev);
error_out:
	kmem_cache_destroy(dazukofs_ign_cachep);
	return err;
}


void dazukofs_ign_dev_destroy(int dev_major, int dev_minor,
			      struct class *dazukofs_class)
{
	class_device_unregister(&ign_dev.cls_dev);
	device_unregister(&ign_dev.dev);
	cdev_del(&ign_dev.cdev);
	dazukofs_destroy_ignlist();
	kmem_cache_destroy(dazukofs_ign_cachep);
}
